@@ -2,15 +2,18 @@
 Timeseries models share a number of common characteristics. This module implements these in a common base class.
 """
 from collections import namedtuple
+import copy
 from copy import deepcopy
 import inspect
 from typing import Any, Callable, Dict, Iterable, List, Tuple, Union
+import warnings
 
 import matplotlib.pyplot as plt
 import numpy as np
 from numpy.lib.function_base import iterable
 import pandas as pd
 from pytorch_lightning import LightningModule
+from pytorch_lightning.trainer.states import RunningStage
 from pytorch_lightning.utilities.parsing import AttributeDict, get_init_args
 import scipy.stats
 import torch
@@ -19,6 +22,7 @@ from torch.nn.utils import rnn
 from torch.optim.lr_scheduler import LambdaLR, ReduceLROnPlateau
 from torch.utils.data import DataLoader
 from tqdm.autonotebook import tqdm
+import yaml
 
 from pytorch_forecasting.data import TimeSeriesDataSet
 from pytorch_forecasting.data.encoders import EncoderNormalizer, GroupNormalizer, MultiNormalizer, NaNLabelEncoder
@@ -31,10 +35,13 @@ from pytorch_forecasting.metrics import (
     MultiHorizonMetric,
     MultiLoss,
     QuantileLoss,
+    convert_torchmetric_to_pytorch_forecasting_metric,
 )
+from pytorch_forecasting.models.nn.embeddings import MultiEmbedding
 from pytorch_forecasting.optim import Ranger
 from pytorch_forecasting.utils import (
     OutputMixIn,
+    TupleOutputMixIn,
     apply_to_list,
     create_mask,
     get_embedding_size,
@@ -69,7 +76,24 @@ def _torch_cat_na(x: List[torch.Tensor]) -> torch.Tensor:
                 )
                 for xi in x
             ]
-    return torch.cat(x, dim=0)
+
+    # check if remaining dimensions are all equal
+    if x[0].ndim > 2:
+        remaining_dimensions_equal = all([all([xi.size(i) == x[0].size(i) for xi in x]) for i in range(2, x[0].ndim)])
+    else:
+        remaining_dimensions_equal = True
+
+    # deaggregate
+    if remaining_dimensions_equal:
+        return torch.cat(x, dim=0)
+    else:
+        # make list instead but warn
+        warnings.warn(
+            f"Not all dimensions are equal for tensors shapes. Example tensor {x[0].shape}. "
+            "Returning list instead of torch.Tensor.",
+            UserWarning,
+        )
+        return [xii for xi in x for xii in xi]
 
 
 def _concatenate_output(
@@ -112,10 +136,22 @@ def _concatenate_output(
                     output_cat[name] = [item for out in output for item in out[name]]
                 else:
                     output_cat[name] = [out[name] for out in output]
+
+    if isinstance(output[0], OutputMixIn):
+        output_cat = output[0].__class__(**output_cat)
     return output_cat
 
 
-class BaseModel(LightningModule):
+STAGE_STATES = {
+    RunningStage.TRAINING: "train",
+    RunningStage.VALIDATING: "val",
+    RunningStage.TESTING: "test",
+    RunningStage.PREDICTING: "predict",
+    RunningStage.SANITY_CHECKING: "sanity_check",
+}
+
+
+class BaseModel(LightningModule, TupleOutputMixIn):
     """
     BaseModel from which new timeseries models should inherit from.
     The ``hparams`` of the created object will default to the parameters indicated in :py:meth:`~__init__`.
@@ -165,6 +201,8 @@ class BaseModel(LightningModule):
 
     """
 
+    CHECKPOINT_HYPER_PARAMS_SPECIAL_KEY = "__special_save__"
+
     def __init__(
         self,
         log_interval: Union[int, float] = -1,
@@ -174,6 +212,7 @@ class BaseModel(LightningModule):
         loss: Metric = SMAPE(),
         logging_metrics: nn.ModuleList = nn.ModuleList([]),
         reduce_on_plateau_patience: int = 1000,
+        reduce_on_plateau_reduction: float = 2.0,
         reduce_on_plateau_min_lr: float = 1e-5,
         weight_decay: float = 0.0,
         optimizer_params: Dict[str, Any] = None,
@@ -197,6 +236,7 @@ class BaseModel(LightningModule):
                 Defaults to [].
             reduce_on_plateau_patience (int): patience after which learning rate is reduced by a factor of 10. Defaults
                 to 1000
+            reduce_on_plateau_reduction (float): reduction in learning rate when encountering plateau. Defaults to 2.0.
             reduce_on_plateau_min_lr (float): minimum learning rate for reduce on plateua learning rate scheduler.
                 Defaults to 1e-5
             weight_decay (float): weight decay. Defaults to 0.0.
@@ -209,13 +249,17 @@ class BaseModel(LightningModule):
             output_transformer (Callable): transformer that takes network output and transforms it to prediction space.
                 Defaults to None which is equivalent to ``lambda out: out["prediction"]``.
             optimizer (str): Optimizer, "ranger", "sgd", "adam", "adamw" or class name of optimizer in ``torch.optim``.
+                Alternatively, a class or function can be passed which takes parameters as first argument and
+                a `lr` argument (optionally also `weight_decay`)
                 Defaults to "ranger".
         """
         super().__init__()
         # update hparams
         frame = inspect.currentframe()
         init_args = get_init_args(frame)
-        self.save_hyperparameters({name: val for name, val in init_args.items() if name not in self.hparams})
+        self.save_hyperparameters(
+            {name: val for name, val in init_args.items() if name not in self.hparams and name not in ["self"]}
+        )
 
         # update log interval if not defined
         if self.hparams.log_val_interval is None:
@@ -223,13 +267,42 @@ class BaseModel(LightningModule):
 
         if not hasattr(self, "loss"):
             if isinstance(loss, (tuple, list)):
-                self.loss = MultiLoss(metrics=loss)
+                self.loss = MultiLoss(metrics=[convert_torchmetric_to_pytorch_forecasting_metric(l) for l in loss])
             else:
-                self.loss = loss
+                self.loss = convert_torchmetric_to_pytorch_forecasting_metric(loss)
         if not hasattr(self, "logging_metrics"):
-            self.logging_metrics = nn.ModuleList([l for l in logging_metrics])
+            self.logging_metrics = nn.ModuleList(
+                [convert_torchmetric_to_pytorch_forecasting_metric(l) for l in logging_metrics]
+            )
         if not hasattr(self, "output_transformer"):
             self.output_transformer = output_transformer
+        if not hasattr(self, "optimizer"):  # callables are removed from hyperparameters, so better to save them
+            self.optimizer = self.hparams.optimizer
+
+        # delete everything from hparams that cannot be serialized with yaml.dump
+        # which is particularly important for tensorboard logging
+        hparams_to_delete = []
+        for k, v in self.hparams.items():
+            try:
+                yaml.dump(v)
+            except:  # noqa
+                hparams_to_delete.append(k)
+                if not hasattr(self, k):
+                    setattr(self, k, v)
+
+        self.hparams_special = getattr(self, "hparams_special", [])
+        self.hparams_special.extend(hparams_to_delete)
+        for k in hparams_to_delete:
+            del self._hparams[k]
+            del self._hparams_initial[k]
+
+    @property
+    def current_stage(self) -> str:
+        """
+        Available inside lightning loops.
+        :return: current trainer stage. One of ["train", "val", "test", "predict", "sanity_check"]
+        """
+        return STAGE_STATES[self.trainer.state.stage]
 
     @property
     def n_targets(self) -> int:
@@ -341,6 +414,18 @@ class BaseModel(LightningModule):
         log.update(self.create_log(x, y, out, batch_idx))
         return log
 
+    def validation_epoch_end(self, outputs):
+        self.epoch_end(outputs)
+
+    def test_step(self, batch, batch_idx):
+        x, y = batch
+        log, out = self.step(x, y, batch_idx)
+        log.update(self.create_log(x, y, out, batch_idx))
+        return log
+
+    def test_epoch_end(self, outputs):
+        self.epoch_end(outputs)
+
     def create_log(
         self,
         x: Dict[str, torch.Tensor],
@@ -374,9 +459,6 @@ class BaseModel(LightningModule):
             )
         return {}
 
-    def validation_epoch_end(self, outputs):
-        self.epoch_end(outputs)
-
     def step(
         self, x: Dict[str, torch.Tensor], y: Tuple[torch.Tensor, torch.Tensor], batch_idx: int, **kwargs
     ) -> Tuple[Dict[str, torch.Tensor], Dict[str, torch.Tensor]]:
@@ -429,13 +511,16 @@ class BaseModel(LightningModule):
             gradient = 0
             # todo: should monotone constrains be applicable to certain targets?
             for pred in prediction_list:
-                gradient = gradient + torch.autograd.grad(
-                    outputs=pred,
-                    inputs=x["decoder_cont"],
-                    grad_outputs=torch.ones_like(pred),  # t
-                    create_graph=True,  # allows usage in graph
-                    allow_unused=True,
-                )[0]
+                gradient = (
+                    gradient
+                    + torch.autograd.grad(
+                        outputs=pred,
+                        inputs=x["decoder_cont"],
+                        grad_outputs=torch.ones_like(pred),  # t
+                        create_graph=True,  # allows usage in graph
+                        allow_unused=True,
+                    )[0]
+                )
 
             # select relevant features
             indices = torch.tensor(
@@ -450,7 +535,7 @@ class BaseModel(LightningModule):
             # multiply monotinicity loss by large number to ensure relevance and take to the power of 2
             # for smoothness of loss function
             monotinicity_loss = 10 * torch.pow(monotinicity_loss, 2)
-            if isinstance(self.loss, MASE):
+            if isinstance(self.loss, (MASE, MultiLoss)):
                 loss = self.loss(
                     prediction, y, encoder_target=x["encoder_target"], encoder_lengths=x["encoder_lengths"]
                 )
@@ -463,14 +548,20 @@ class BaseModel(LightningModule):
 
             # calculate loss
             prediction = out["prediction"]
-            if isinstance(self.loss, MASE):
-                loss = self.loss(
-                    prediction, y, encoder_target=x["encoder_target"], encoder_lengths=x["encoder_lengths"]
-                )
+            if isinstance(self.loss, (MASE, MultiLoss)):
+                mase_kwargs = dict(encoder_target=x["encoder_target"], encoder_lengths=x["encoder_lengths"])
+                loss = self.loss(prediction, y, **mase_kwargs)
             else:
                 loss = self.loss(prediction, y)
 
-        self.log(f"{['val', 'train'][self.training]}_loss", loss, on_step=self.training, on_epoch=True, prog_bar=True)
+        self.log(
+            f"{self.current_stage}_loss",
+            loss,
+            on_step=self.training,
+            on_epoch=True,
+            prog_bar=True,
+            batch_size=len(x["decoder_target"]),
+        )
         log = {"loss": loss, "n_samples": x["decoder_lengths"].size(0)}
         return log, out
 
@@ -514,37 +605,17 @@ class BaseModel(LightningModule):
                 else:
                     loss_value = metric(y_point, y_true)
                 if len(y_hat_point_detached) > 1:
-                    target_tag = f"Target {idx}_"
+                    target_tag = self.target_names[idx] + " "
                 else:
                     target_tag = ""
                 self.log(
-                    f"{target_tag}{['val', 'train'][self.training]}_{metric.name}",
+                    f"{target_tag}{self.current_stage}_{metric.name}",
                     loss_value,
                     on_step=self.training,
                     on_epoch=True,
+                    batch_size=len(x["decoder_target"]),
                 )
 
-    def to_network_output(self, **results):
-        """
-        Convert output into a named (and immuatable) tuple.
-
-        This allows tracing the modules as graphs and prevents modifying the output.
-
-        Returns:
-            named tuple
-        """
-        if hasattr(self, "_output_class"):
-            Output = self._output_class
-        else:
-            OutputTuple = namedtuple("output", results)
-
-            class Output(OutputMixIn, OutputTuple):
-                pass
-
-            self._output_class = Output
-
-        return self._output_class(**results)
-
     def forward(
         self, x: Dict[str, Union[torch.Tensor, List[torch.Tensor]]]
     ) -> Dict[str, Union[torch.Tensor, List[torch.Tensor]]]:
@@ -620,7 +691,7 @@ class BaseModel(LightningModule):
             x (Dict[str, torch.Tensor]): x as passed to the network by the dataloader
             out (Dict[str, torch.Tensor]): output of the network
             batch_idx (int): current batch index
-            **kwargs: paramters to pass to ``pot_prediction``
+            **kwargs: paramters to pass to ``plot_prediction``
         """
         # log single prediction figure
         if (batch_idx % self.log_interval == 0 or self.log_interval < 1.0) and self.log_interval > 0:
@@ -632,7 +703,7 @@ class BaseModel(LightningModule):
                 log_indices = [0]
             for idx in log_indices:
                 fig = self.plot_prediction(x, out, idx=idx, add_loss_to_title=True, **kwargs)
-                tag = f"{['Val', 'Train'][self.training]} prediction"
+                tag = f"{self.current_stage} prediction"
                 if self.training:
                     tag += f" of item {idx} in global batch {self.global_step}"
                 else:
@@ -640,7 +711,7 @@ class BaseModel(LightningModule):
                 if isinstance(fig, (list, tuple)):
                     for idx, f in enumerate(fig):
                         self.logger.experiment.add_figure(
-                            f"Target {idx} {tag}",
+                            f"{self.target_names[idx]} {tag}",
                             f,
                             global_step=self.global_step,
                         )
@@ -831,63 +902,72 @@ class BaseModel(LightningModule):
             optimizer_params = {}
         else:
             optimizer_params = self.hparams.optimizer_params
-        if isinstance(self.hparams.learning_rate, (list, tuple)):  # set schedule
-            lrs = self.hparams.learning_rate
-            if self.hparams.optimizer == "adam":
-                optimizer = torch.optim.Adam(self.parameters(), lr=lrs[0])
-            elif self.hparams.optimizer == "adamw":
-                optimizer = torch.optim.AdamW(self.parameters(), lr=lrs[0])
-            elif self.hparams.optimizer == "ranger":
-                optimizer = Ranger(self.parameters(), lr=lrs[0], weight_decay=self.hparams.weight_decay)
-            elif self.hparams.optimizer == "sgd":
-                optimizer = torch.optim.SGD(
-                    self.parameters(), lr=lrs[0], weight_decay=self.hparams.weight_decay, **optimizer_params
+        # set optimizer
+        lrs = self.hparams.learning_rate
+        if isinstance(lrs, (list, tuple)):
+            lr = lrs[0]
+        else:
+            lr = lrs
+        if callable(self.optimizer):
+            try:
+                optimizer = self.optimizer(
+                    self.parameters(), lr=lr, weight_decay=self.hparams.weight_decay, **optimizer_params
                 )
-            else:
-                try:
-                    optimizer = getattr(torch.optim, self.hparams.optimizer)(
-                        self.parameters(), lr=lrs[0], weight_decay=self.hparams.weight_decay, **optimizer_params
-                    )
-                except AttributeError:
-                    raise ValueError(f"Optimizer of self.hparams.optimizer={self.hparams.optimizer} unknown")
+            except TypeError:  # in case there is no weight decay
+                optimizer = self.optimizer(self.parameters(), lr=lr, **optimizer_params)
+        elif self.hparams.optimizer == "adam":
+            optimizer = torch.optim.Adam(
+                self.parameters(), lr=lr, weight_decay=self.hparams.weight_decay, **optimizer_params
+            )
+        elif self.hparams.optimizer == "adamw":
+            optimizer = torch.optim.AdamW(
+                self.parameters(), lr=lr, weight_decay=self.hparams.weight_decay, **optimizer_params
+            )
+        elif self.hparams.optimizer == "ranger":
+            optimizer = Ranger(self.parameters(), lr=lr, weight_decay=self.hparams.weight_decay, **optimizer_params)
+        elif self.hparams.optimizer == "sgd":
+            optimizer = torch.optim.SGD(
+                self.parameters(), lr=lr, weight_decay=self.hparams.weight_decay, **optimizer_params
+            )
+        elif hasattr(torch.optim, self.hparams.optimizer):
+            try:
+                optimizer = getattr(torch.optim, self.hparams.optimizer)(
+                    self.parameters(), lr=lr, weight_decay=self.hparams.weight_decay, **optimizer_params
+                )
+            except TypeError:  # in case there is no weight decay
+                optimizer = getattr(torch.optim, self.hparams.optimizer)(self.parameters(), lr=lr, **optimizer_params)
+        else:
+            raise ValueError(f"Optimizer of self.hparams.optimizer={self.hparams.optimizer} unknown")
+
+        # set scheduler
+        if isinstance(lr, (list, tuple)):  # change for each epoch
             # normalize lrs
             lrs = np.array(lrs) / lrs[0]
-            schedulers = [
-                {
-                    "scheduler": LambdaLR(optimizer, lambda epoch: lrs[min(epoch, len(lrs) - 1)]),
-                    "interval": "epoch",
-                    "reduce_on_plateau": False,
-                    "frequency": 1,
-                }
-            ]
+            scheduler_config = {
+                "scheduler": LambdaLR(optimizer, lambda epoch: lrs[min(epoch, len(lrs) - 1)]),
+                "interval": "epoch",
+                "frequency": 1,
+                "strict": False,
+            }
+        elif self.hparams.reduce_on_plateau_patience is None:
+            scheduler_config = {}
         else:  # find schedule based on validation loss
-            if self.hparams.optimizer == "adam":
-                optimizer = torch.optim.Adam(self.parameters(), lr=self.hparams.learning_rate)
-            elif self.hparams.optimizer == "ranger":
-                optimizer = Ranger(
-                    self.parameters(), lr=self.hparams.learning_rate, weight_decay=self.hparams.weight_decay
-                )
-            elif self.hparams.optimizer == "adamw":
-                optimizer = torch.optim.AdamW(self.parameters(), lr=self.hparams.learning_rate)
-            else:
-                raise ValueError(f"Optimizer of self.hparams.optimizer={self.hparams.optimizer} unknown")
-            schedulers = [
-                {
-                    "scheduler": ReduceLROnPlateau(
-                        optimizer,
-                        mode="min",
-                        factor=0.1,
-                        patience=self.hparams.reduce_on_plateau_patience,
-                        cooldown=self.hparams.reduce_on_plateau_patience,
-                        min_lr=self.hparams.reduce_on_plateau_min_lr,
-                    ),
-                    "monitor": "val_loss",  # Default: val_loss
-                    "interval": "epoch",
-                    "reduce_on_plateau": True,
-                    "frequency": 1,
-                }
-            ]
-        return [optimizer], schedulers
+            scheduler_config = {
+                "scheduler": ReduceLROnPlateau(
+                    optimizer,
+                    mode="min",
+                    factor=1.0 / self.hparams.reduce_on_plateau_reduction,
+                    patience=self.hparams.reduce_on_plateau_patience,
+                    cooldown=self.hparams.reduce_on_plateau_patience,
+                    min_lr=self.hparams.reduce_on_plateau_min_lr,
+                ),
+                "monitor": "val_loss",  # Default: val_loss
+                "interval": "epoch",
+                "frequency": 1,
+                "strict": False,
+            }
+
+        return {"optimizer": optimizer, "lr_scheduler": scheduler_config}
 
     @classmethod
     def from_dataset(cls, dataset: TimeSeriesDataSet, **kwargs) -> LightningModule:
@@ -921,9 +1001,29 @@ class BaseModel(LightningModule):
         )  # add dataset parameters for making fast predictions
         # hyper parameters are passed as arguments directly and not as single dictionary
         checkpoint["hparams_name"] = "kwargs"
+        # save specials
+        checkpoint[self.CHECKPOINT_HYPER_PARAMS_SPECIAL_KEY] = {k: getattr(self, k) for k in self.hparams_special}
+        # add special hparams them back to save the hparams correctly for checkpoint
+        checkpoint[self.CHECKPOINT_HYPER_PARAMS_KEY].update(checkpoint[self.CHECKPOINT_HYPER_PARAMS_SPECIAL_KEY])
+
+    @property
+    def target_names(self) -> List[str]:
+        """
+        List of targets that are predicted.
+
+        Returns:
+            List[str]: list of target names
+        """
+        if hasattr(self, "dataset_parameters") and self.dataset_parameters is not None:
+            return to_list(self.dataset_parameters["target"])
+        else:
+            return [f"Target {idx + 1}" for idx in range(self.n_targets)]
 
     def on_load_checkpoint(self, checkpoint: Dict[str, Any]) -> None:
         self.dataset_parameters = checkpoint.get("dataset_parameters", None)
+        # load specials
+        for k, v in checkpoint[self.CHECKPOINT_HYPER_PARAMS_SPECIAL_KEY].items():
+            setattr(self, k, v)
 
     def to_prediction(self, out: Dict[str, Any], use_metric: bool = True, **kwargs):
         """
@@ -1247,6 +1347,25 @@ class BaseModelWithCovariates(BaseModel):
             as bag of embeddings
     """
 
+    @property
+    def target_positions(self) -> torch.LongTensor:
+        """
+        Positions of target variable(s) in covariates.
+
+        Returns:
+            torch.LongTensor: tensor of positions.
+        """
+        # todo: expand for categorical targets
+        if "target" in self.hparams:
+            target = self.hparams.target
+        else:
+            target = self.dataset_parameters["target"]
+        return torch.tensor(
+            [self.hparams.x_reals.index(name) for name in to_list(target)],
+            device=self.device,
+            dtype=torch.long,
+        )
+
     @property
     def reals(self) -> List[str]:
         """List of all continuous variables in model"""
@@ -1354,6 +1473,47 @@ class BaseModelWithCovariates(BaseModel):
         new_kwargs.update(kwargs)
         return super().from_dataset(dataset, **new_kwargs)
 
+    def extract_features(
+        self,
+        x,
+        embeddings: MultiEmbedding = None,
+        period: str = "all",
+    ) -> torch.Tensor:
+        """
+        Extract features
+
+        Args:
+            x (Dict[str, torch.Tensor]): input from the dataloader
+            embeddings (MultiEmbedding): embeddings for categorical variables
+            period (str, optional): One of "encoder", "decoder" or "all". Defaults to "all".
+
+        Returns:
+            torch.Tensor: tensor with selected variables
+        """
+        # select period
+        if period == "encoder":
+            x_cat = x["encoder_cat"]
+            x_cont = x["encoder_cont"]
+        elif period == "decoder":
+            x_cat = x["decoder_cat"]
+            x_cont = x["decoder_cont"]
+        elif period == "all":
+            x_cat = torch.cat([x["encoder_cat"], x["decoder_cat"]], dim=1)  # concatenate in time dimension
+            x_cont = torch.cat([x["encoder_cont"], x["decoder_cont"]], dim=1)  # concatenate in time dimension
+        else:
+            raise ValueError(f"Unknown type: {type}")
+
+        # create dictionary of encoded vectors
+        input_vectors = embeddings(x_cat)
+        input_vectors.update(
+            {
+                name: x_cont[..., idx].unsqueeze(-1)
+                for idx, name in enumerate(self.hparams.x_reals)
+                if name in self.reals
+            }
+        )
+        return input_vectors
+
     def calculate_prediction_actual_by_variable(
         self,
         x: Dict[str, torch.Tensor],
@@ -1545,7 +1705,7 @@ class BaseModelWithCovariates(BaseModel):
                 x = np.linspace(-data["std"], data["std"], bins)
                 # reversing normalization for group normalizer is not possible without sample level information
                 if not isinstance(scaler, (GroupNormalizer, EncoderNormalizer)):
-                    x = scaler.inverse_transform(x)
+                    x = scaler.inverse_transform(x.reshape(-1, 1)).reshape(-1)
                     ax.set_xlabel(f"Normalized {name}")
 
                 if len(x) > 0:
@@ -1883,21 +2043,6 @@ class AutoRegressiveBaseModelWithCovariates(BaseModelWithCovariates, AutoRegress
             as bag of embeddings
     """
 
-    @property
-    def target_positions(self) -> torch.LongTensor:
-        """
-        Positions of target variable(s) in covariates.
-
-        Returns:
-            torch.LongTensor: tensor of positions.
-        """
-        # todo: expand for categorical targets
-        return torch.tensor(
-            [self.hparams.x_reals.index(name) for name in to_list(self.hparams.target)],
-            device=self.device,
-            dtype=torch.long,
-        )
-
     @property
     def lagged_target_positions(self) -> Dict[int, torch.LongTensor]:
         """
